
\ eforth.f v0.6.1b (c) Васильев И.В., Москва, 2019
\ 86eForth502.asm (c) д-р Чэнь-Хэнсон Тинг, 1990-2016
\ EFORTH.X86 (c) Билл Мунч, Санта-Круз, Калифорния, 1989-1999

\ Оболочка eForth: область данных и определения высокого уровня
\ Исходный текст написан 24 ноября 2019 для SP-Forth 4.20

\ На основе кода 86eforth502.asm и книги доктора Тинга
\ EForth and Zen, 3rd Edition 2017

ONLY FORTH DEFINITIONS
VOCABULARY VMETA                \ словарь кросс-компилятора
VOCABULARY DEFS                 \ слова и метки целевой машины

ALSO VMETA DEFINITIONS          \ новые определения в мете, держим FORTH нетронутым
: {  ( -- )  ONLY FORTH ALSO VMETA ALSO DEFS DEFINITIONS ;  \ создавать слова в таблице символов
: }  ( -- )  ONLY FORTH ALSO VMETA DEFINITIONS ;            \ определять слова метаязыка

   8 CONSTANT BKSPP              \ литера: пробел назад
   9 CONSTANT TABB               \ литера: табуляция
  13 CONSTANT CRR                \ литера: возврат каретки

   4 CONSTANT COMPO              \ лексикон: бит только для компиляции
   8 CONSTANT IMEDD              \ лексикон: бит немедленого исполнения

\ Переменные драйверов МК-161 и другие регистры, для последующей инициализации
  20 CONSTANT ptrKbdInt         \ состояние клавиатуры

  41 CONSTANT rINTER            \ вектор $INTERPRET для [ в формате 3+3
  42 CONSTANT rCOMP             \ вектор $COMPILE для ] в формате 3+3
  43 CONSTANT rCOLD             \ СВУ, которое ядро запускает первым

5093 CONSTANT flgKbdCaps        \ флаг клавиатуры: строчные / заглавные = 0 / 1

\ X-компилятор: вспомогательные слова
   0 VALUE wc                   \ счётчик откомпилированных слов
   0 VALUE mc                   \ счётчик примитивов

: wc++  ( -- )   wc 1+ TO wc ;  \ подсчёт СВУ
: mc++  ( -- )   mc 1+ TO mc ;  \ подсчёт примитивов

: WBin@ ( адрес -- n )                          \ считать 2-х байтовое целое без знака в двоичном формате ЭКВМ
    DUP C@ 256 * SWAP 1+ C@ + ;
: WBin! ( n адрес -- )  OVER 256 / OVER C!      \ записать 2-х байтовое целое без знака в двоичном формате ЭКВМ
    1+ SWAP 256 MOD SWAP C! ;
: WA!   ( a адрес -- )  OVER 100 / OVER C!      \ записать 2-х байтовое целое без знака в двоичном формате ЭКВМ (mod 100)
    1+ SWAP 100 MOD SWAP C! ;
: C>BCD ( c -- n )      10 /MOD 16 * + ;        \ перевести байт c в BCD
: BCD>C ( n -- c )      16 /MOD 10 * + ;        \ перевести байт n из BCD
: WBCD@ ( адрес -- n )
    DUP C@ BCD>C 100 *  SWAP 1+ C@ BCD>C + ;    \ считать 2-х байтовое целое число без знака в bcd-формате ЭКВМ
: WBCD! ( n адрес -- ) OVER 100 / C>BCD OVER C! \ записать 2-х байтовое целое число без знака в bcd-формате ЭКВМ
    1+ SWAP 100 MOD C>BCD SWAP C! ;
: >3+3  ( a -- n )     256 /MOD 1000 * + ;      \ перевести адрес в формат 3+3

\ Шпаргалка по внутреннему формату МК-161:

\ -name code лекс 11 поля заголовка слова, name=nfa, code=cfa=xt
\ +nnnn cccc llll 11 заголовок слова
\ +reg0 maxr used 11 поля заголовка списка
\ sAAAA BBBB CCCC 11 адресация TRA!, TRB!, TRC! и TRC@
\ 00000 0000 0111    номер цифры (ст)
\ 01234 5678 9012    номер цифры (мл)
\ 01122 3344 5566 77 смещение байта (0-7)

: DecShiftRight ( a -- )              \ сдвинуть мантиссу вправо, освободив место для новой цифры слева
  7 OVER + DUP C@ 1+ SWAP C!          \ увеличить порядок на 1
  0                                   \ вставить цифру 0
  6 0 DO                ( a цифра )   \ цикл выполнится 6 раз
    16 * SWAP 1+ SWAP   ( a+1 16*цифра )
    OVER C@ 16 /MOD     ( a+1 16*цифра мл ст )
    ROT +               ( a+1 мл новбайт )
    2 PICK C!           ( a+1 мл )
  LOOP
  2DROP ;

: DecShiftLeft ( a -- )               \ сдвинуть мантиссу влево, уничтожив старший разряд
  7 + DUP C@ 1- OVER C! ( a+7 )       \ уменьшить порядок на 1
  0                     ( a+7 0 )     \ вставить цифру 0, начнём с конца мантиссы
  6 0 DO                ( a цифра )   \ цикл выполнится 6 раз
    SWAP 1- SWAP        ( a-1 цифра )
    OVER C@ 16 /MOD     ( a-1 цифра мл ст )
    ROT ROT 16 * +      ( a-1 ст новбайт )
    2 PICK C!           ( a-1 ст )
  LOOP
  2DROP ;

: Norm ( a -- )                       \ нормализовать десятичное число
  DUP 0                 ( a a 0 )     \ начать с нуля
  6 0 DO                ( a a x )     \ цикл выполнится 6 раз
    SWAP 1+ SWAP        ( a a+1 x )
    OVER C@ OR          ( a a+1 x )
  LOOP
  SWAP DROP
  IF BEGIN DUP 1+ C@ 16 / 0= WHILE \ пока старшая цифра ноль,
       DUP DecShiftLeft     \ сдвигать мантиссу влево
     REPEAT
     DROP
  ELSE                      \ если вся мантисса 0,
    0 OVER C!               \ тогда и знак ноль
    7 + 0 SWAP C!           \ и порядок ноль
  THEN ;

: StoreDigit ( a цифра -- ) \ добавить старшую десятичную цифру к числу во внутреннем формате МК-161
    16 * SWAP 1+ DUP C@ ( d*16 a+1 byte )
    ROT + SWAP C! ;

: dec! ( n a -- )           \ записать целое n во внутреннем формате МК-161
  DUP 8 ERASE               \ обнулить 8-байтный буфер
  SWAP DUP 0< IF 0x80 2 PICK C! NEGATE THEN \ записать знак числа
  ( a n )
  BEGIN 10 /MOD         ( a цифра n' )
    SWAP 2 PICK SWAP    ( a n' a цифра )
    StoreDigit          ( a n' )
    DUP
  WHILE OVER DecShiftRight
  REPEAT
  2DROP ;


\ Работа с целевой двоичной памятью
CREATE regb  8168 ALLOT     \ регистровая память (байты)
regb 8168 ERASE             \ обнулить

: save.mkb ( -- )           \ сохранить байтовые регистры в файл eforth.mkb
  S" eforth.mkb" W/O ( BIN )
  CREATE-FILE ABORT" Can't create eforth.mkb"
  1000 regb + 4096 2 PICK
  WRITE-FILE  ABORT" Can't write MKB"
  CLOSE-FILE  ABORT" Can't close MKB"
;

VARIABLE CP'  1000 CP' !              \ вершина целевого словаря
: THERE   ( reg -- a )   regb + ;     \ отобразить целевой адрес в инструментальный, для байтовой памяти
: HERE'   ( -- reg )     CP' @  ;     \ вершина целевого словаря
: ALLOT'  ( n -- )       CP' +! ;     \ зарезервировать n байтовых регистров цели
: TC@     ( reg -- c )   THERE C@ ;   \ считать байтовый регистр цели
: TC!     ( c reg -- )   THERE C! ;   \ записать байтовый регистр цели
: T@      ( reg -- n )   THERE WBin@ ; \ считать целое из байтовой памяти цели, старший байт идёт первым
: T!      ( n reg -- )   THERE WBin! ; \ записать целое n в два целевых байтовых регистра
: Txt!    ( xt reg -- )  THERE WA! ;  \ записать токен xt в два целевых байтовых регистра (mod 100)
: TC,     ( c -- )       HERE' TC! 1 ALLOT' ; \ скомпилировать байт c в байтовую память цели
: T,      ( n -- )       256 /MOD TC, TC, ; \ скомпилировать целое n в байтовую память цели
: Txt,    ( xt -- )      100 /MOD TC, TC, ; \ скомпилировать токен xt в байтовую память цели (mod 100)
: T,"     ( "str" -- )   34 WORD COUNT SWAP \ ( # from ), скомпилировать строку в бпц
    HERE' THERE 2 PICK MOVE ALLOT' ;


VARIABLE UP'  5082 UP' !        \ область системных переменных
: UALLOT' ( n -- )  NEGATE UP' +! ; \ зарезервировать n байт в области USER
: Label'  ( "name" -- )         \ создать инструментальную метку на целевую байтовую память
  { CREATE } UP' @ ,            \ инструментальное слово содержит целевой адрес
  DOES>   ( -- n )  @           \ при употреблении метка кладёт свой целевой адрес на стек
;

: =db     ( n "name" -- )       \ создать системную переменную с начальным значением n
  1 UALLOT'                     \ зарезервировать 1 байт в области USER
  UP' @ TC!                     \ инициализировать системную переменную её начальным значением
  Label'                        \ создать инструментальное слово-метку
;

: =dw     ( n "name" -- )       \ создать системную переменную с начальным значением n
  2 UALLOT'                     \ зарезервировать 2 байта в области USER
  UP' @ T!                      \ инициализировать системную переменную её начальным значением
  Label'                        \ создать инструментальное слово-метку
;

: .dw   ( "name" -- )   0 =dw ; \ создать системную переменную с начальным значением 0
: .db   ( "name" -- )   0 =db ; \ создать системную переменную с начальным значением 0


\ Системные переменные eForth, от старших адресов к младшим
	.dw _LASTN              \ адрес заголовка слова, определённого последним
	.dw _CP                 \ верх словаря, ниже инициализируется HERE'
	.dw _TBOOT              \ вектор слова загрузки
	.dw _HLD                \ следующая литера в строке числового вывода
	.dw _EVAL               \ вектор исполнения для EVAL
	.dw _TIBB               \ начало строки ввода, ниже инициализируется _TIB
	.dw _NTIB               \ длина строки ввода
 30	=db _SIGMA              \ при первом запуске Sigma=30
	.db _SIGMA1             \ *** убрать
	.dw _IN                 \ указатель на текущую литеру в строке ввода
255	=dw _8BIT               \ 127/255, выкл/вкл литеры 128..255
	.dw _BASE               \ основание системы счисления для числового ввода/вывода
1000	=dw _RPP                \ сохранить дно стека возвратов (R2) для QUIT
 898	=dw _SPP                \ сохранить дно стека данных (R3) для ABORT


80 UALLOT' Label' _TIB          \ буфер ввода терминала сразу под областью USER


\ Стандартные сообщения TYPE
: MSG_PROMPT  ( -- )   1 TC, 0 TC, ; \ " > "
: MSG_UNDERF  ( -- )   1 TC, 1 TC, ; \ " underflow "
: MSG_COMPO   ( -- )   1 TC, 2 TC, ; \ " compile only"
: MSG_REDEF   ( -- )   1 TC, 3 TC, ; \ " reDef "
: MSG_NAME    ( -- )   1 TC, 4 TC, ; \ " name"
: MSG_NONAME  ( -- )   1 TC, 5 TC, ; \ " {noName}"
: MSG_SIGNON  ( -- )   1 TC, 6 TC, ; \ "161eForth v#.#"
: MSG_BKSP    ( -- )   1 TC, 7 TC, ; \ ^h BL BL ^h ^h


\ Работа с целевой десятичной памятью
VARIABLE nNames                 \ количество зашитых имён

CREATE regd  8000 ALLOT         \ регистровая память (десятичные регистры)
regd 8000 ERASE                 \ обнулить

: RTHERE   ( reg -- a ) 8 * regd + ;  \ дать инструментальный адрес десятичного регистра цели
: TR!      ( n reg -- ) RTHERE dec! ; \ присвоить целое значение десятичному регистру

: TRB!  ( n reg -- )   RTHERE 3 + WBCD! ; \ записать средние 4 цифры десятичного регистра цели (CFA)
: TRA!  ( n reg -- )            \ записать старшие 4 цифры десятичного регистра цели (NFA)
   RTHERE 11 OVER 7 + C!        \ порядок (07) := 11
   1+ WBCD!  ;                  \ первые 4 знака мантиссы n (01-02)

: TRC@  ( reg -- )     RTHERE 5 + WBCD@ ; \ считать младшие 4 цифры десятичного регистра цели (байт лексикона)
: TRC!  ( c reg -- )   RTHERE 5 + WBCD! ; \ записать младшие 4 цифры десятичного регистра цели (байт лексикона)

: TRSetNeg ( reg -- )  RTHERE 128 SWAP C! ; \ сделать число в десятичном регистре цели отрицательным

: NormAll ( -- )                \ нормализовать все десятичные регистры
  1000 0 DO I RTHERE Norm LOOP  \ цикл от 0 до 999, тысячу раз
;

: save.mkd ( -- )               \ сохранить десятичные регистры в файл eforth.mkd
  333                           \ список слов пользователя длиной в 1 литеру
  333 301 DO                    \ I от 301 до 332, для слов длиной от 1 до 32
    DUP I TRA!                  \ A: адрес первого регистра
    2 I TRB!                    \ B:2 выделить 2 регистра C:0 использованных нет
    2+                          \ резерв двух регистров
  LOOP                          \ инициализация следующего списка слов
  DROP                          \ убрать номер регистра из стека
  NormAll                       \ подчистить старший ноль
  S" eforth.mkd" W/O ( BIN )
  CREATE-FILE ABORT" Can't create eforth.mkd"
  regd 8000 2 PICK
  WRITE-FILE  ABORT" Can't write MKD"
  CLOSE-FILE  ABORT" Can't close MKD"
;


\ Работа с целевой памятью программ
CREATE prgm  10000 ALLOT        \ память программ
prgm 10000 255 FILL             \ заполнить FF
0 VALUE nPages                  \ количество программных страниц в eforth

\ Заголовок ядра содержит постоянные, зависящие от сборки
VARIABLE tblNames		\ таблица строк, отрезать всё ниже неё
VARIABLE tblTokens		\ таблица адресов примитивов
VARIABLE &KPNum			\ режим клавиатуры по умолчанию
VARIABLE INEXT'			\ INEXT: внутренний интерпретатор
VARIABLE doLst'			\ DOLST: обработчик двоеточия
VARIABLE doVar'			\ DOVAR: обработчик VARIABLE и CREATE
VARIABLE doCon'			\ DOCON: обработчик положительных CONSTANT
VARIABLE doConM'		\ DOCONM: обаботчик отрицательных CONSTANT

: ITHERE   ( n -- a )  prgm +  ;      \ отобразить целевой адрес в инструментальный, для памяти программ
: TIC!     ( c a -- )  ITHERE C! ;    \ записать шаг c по адресу a в целевую память программ
: TIC@     ( a -- c )  ITHERE C@ ;    \ считать шаг c по адресу a из целевой памяти программ
: TI!      ( n a -- )  ITHERE WBin! ; \ записать двоичное целое n в два шага цели по адресу a
: TI@      ( a -- n )  ITHERE WBin@ ; \ считать двоичное целое n из двух шагов цели по адресу a
: TIbcd!   ( c a -- )  ITHERE  SWAP 100 MOD C>BCD SWAP C! ; \ записать шаг c по адресу a в целевую память программ, как число bcd
: TIbcd@   ( a -- c )  ITHERE  C@ BCD>C ; \ считать десятичный шаг c из адреса a целевой памяти программ
: TIWbcd!  ( n a -- )  ITHERE  WBCD! ; \ записать целое n в два шага по адресу a, как число bcd
: TIWbcd@  ( a -- n )  ITHERE  WBCD@ ; \ считать десятичное целое n из двух шагов целевой памяти программ

: load.mkp ( -- )               \ загрузить заготовку ядра eForth
  S" eforth0.mkp" R/O ( BIN )
  OPEN-FILE ABORT" Can't open eforth0.mkp"  ( fileid )
  DUP FILE-SIZE ABORT" Can't get MKP size"  ( fileid udSize )
  ABORT" MKP is too big"                    ( fileid uSize )               \ размер MKP должен быть <= 20001 байт, тем более <64Kбайт
  HERE OVER ALLOT                           ( fileid uSize ptrImage )      \ выделить память под образ MKP
  2DUP SWAP 4 PICK                          ( fileid uSize ptrImage ptrImage uSize fileid)
  READ-FILE ABORT" Can't read MKP"          ( fileid uSize ptrImage uRead )
\ 198 OVER + 200 / TO nPages    \ число загруженных страниц
  2/ 0 DO 1+                      ( fileid uSize ptrImage+1 )
    DUP C@  I TIC!              \ копировать считанные шаги в целевую память программ
    1+                          \ следующий шаг в образе -- через байт
  LOOP DROP                     \ скопировать память программ из образа MKP
  NEGATE ALLOT                  \ убрать образ из памяти
  CLOSE-FILE  ABORT" Can't close MKP"

\ Предобработка: считать заголовок ядра
  1 TIbcd@                                     \ найти адрес MAIN и заголовка
  2- DUP TI@ 2019 - ABORT" Wrong magic number" \ проверить почтовую марку
  2- DUP TI@ tblNames !
  2- DUP TI@ tblTokens !
  2- DUP TI@ &KPNum !
  1- DUP TIC@ INEXT' !
  1- DUP TIC@ doLst' !
  1- DUP TIC@ doVar' !
  1- DUP TIC@ doCon' !
  1- DUP TIC@ doConM' ! ;

: save.mkp ( -- )               \ сохранить ядро eForth в файл eforth.mkp
  HERE 0 C,                     \ номер начальной страницы: 0
  prgm nPages 100 * 0 DO DUP C@ C, 1 C, 1+ LOOP DROP \ завершить образ MKP
  S" eforth.mkp" W/O ( BIN )
  CREATE-FILE ABORT" Can't create eforth.mkp"
  SWAP nPages 200 * 1+ 2 PICK
  WRITE-FILE  ABORT" Can't write MKP"
  CLOSE-FILE  ABORT" Can't close MKP"
  nPages 200 * 1+ NEGATE ALLOT ; \ убрать образ из памяти

: process.mkp ( -- )            \ обработать MKP
  nNames @ 0 DO                 \ построить таблицу заголовков
    tblNames @ I 2* + TI@       \ получить адрес имени
    I 44 + DUP TRSetNeg TRA!    \ заголовок отрицательный, записать NFA
    tblTokens @ I 2* + TI@      \ получить адрес примитива
    DUP IF I DUP 44 + TRB! THEN \ если не 0, записать его xt в CFA
  LOOP
  tblTokens @  nNames @  2* +   \ адрес, с которого добить таблицу нулями
  256 nNames @ - 0 DO           \ количество токенов, требующих инциализации
    0 OVER TI!                  \ записать 0 в адрес токена
    2+                          \ перейти к следующему токену
  LOOP
  BEGIN  DUP 100 MOD            \ страница закончилась?
  WHILE  255 OVER TIC!  1+      \ забить шаг FF, перейти к следующему
  REPEAT                        \ последняя страница забита нулями
  100 / TO nPages ;             \ сохранить число страниц ядра

load.mkp                        \ загрузить заготовку ядра


\ X-компилятор
VARIABLE xtEXIT                 \ токен примитива EXIT
VARIABLE xtBRAN                 \ токен примитива branch
VARIABLE xtQBRAN                \ токен примитива ?branch

VARIABLE Last'                  \ номер последнего X-определённого слова
VARIABLE flgOpt                 \ флаг, возможно ли оптимизировать хвостовую рекурсию

: SetLex  ( c -- )              \ установить биты c в байте лексикона последнего слова
  Last' @ 44 + TRC@ OR          \ считать лексикон, установить запрошенные биты
  Last' @ 44 + TRC! ;           \ сохранить лексикон

: ::      ( n "имя" -- )        \ определить СВУ слова n в целевом словаре
  DUP Last' !                   \ сохранить номер слова
  ?DUP IF 44 + HERE' SWAP TRB! THEN \ если n#0, записать CFA в поле кода регистра заголовка
  { CREATE HERE' ,              \ создать инструментальное слово, сохранить целевое CFA
  doLst' @ TC,                  \ записать обработчик DOLST в поле кода
  DOES>   ( -- )  @ Txt,        \ при исполнении слово компилирует своё CFA, как токен
  1 flgOpt ! ;                  \ для СВУ оптимизация ; возможна

\ *** позже можно переименовать в ;0 и ;
: ;;      ( -- )  } wc++ ;      \ завершить слово высокого уровня без EXIT в ABORT и QUIT

: ;x      ( -- ) xtEXIT @ T, } wc++ ; \ завершить слово с компиляцией EXIT

\ Очевидные ошибки оптимизации выявляются, но бывают и сложные случаи.
: ;-      ( -- )                \ завершить слово с оптимизацией хвостовой рекурсии
  flgOpt @ 0= ABORT" Optimization failed" \ грубая проверка, возможна ли оптимизация
  HERE' 2- DUP T@               \ считать последний токен, можно устроить проверку на СВУ
  xtBRAN @ ROT T!               \ записать на его место BRANCH
  T,                            \ скомпилировать токен, как адрес BRANCH (pfa-1 = cfa)
  } wc++ ;                      \ ещё одно СВУ успешно скомпилировано

: IMMEDIATE     ( -- )   IMEDD SetLex ; \ установить флаг IMMEDIATE последнего целевого слова
: COMPILE-ONLY  ( -- )   COMPO SetLex ; \ установить флаг COMPILE-ONLY последнего целевого слова

: b:      ( -- )
  { CREATE HERE' ,              \ создать инструментальное слово, сохранив текущий адрес
  DOES>   ( -- )  @ 1- Txt, ;   \ при исполнении компилирует адрес перехода туда

: prime   ( n "имя" -- )        \ объявить примитив
  DUP Last' !                   \ сохранить номер примитива
  { CREATE }  C,  mc++          \ для работы IMMEDIATE и COMP-ONLY
  DOES>   ( -- )  0 TC, C@ TC,  \ при исполнении примитив компилирует свой номер
  0 flgOpt ! ;                  \ оптимизация ; невозможна

: high    ( n "имя" -- )        \ объявить СВУ
  DUP Last' !                   \ сохранить номер СВУ
  } CREATE C,                   \ для работы IMMEDIATE и COMP-ONLY
  DOES>    ( -- )  C@ ;         \ получить номер СВУ

: [F      ( -- )   } CREATE HERE' { ,     \ определить область памяти, как ссылку вперёд
  DOES>   ( -- a ) HERE' SWAP @ CP' ! { ; \ при исполнении ссылка вперёд восстанавливает место компиляции

: F]      ( a -- ) CP' ! } ;    \ возвратить место компиляции, после разрешения ссылки вперёд


\ Структуры управления X-компилятора
\ *** возможно, их требуется значительно меньше
\ *** можно добавить глубокую оптимизацию, когда ;- после THEN изменяет адрес ?BRANCH
\ *** ещё одна возможная оптимизация, когда ; не компилирует EXIT в QUIT после AGAIN
: BEGIN   ( -- a )     HERE'                    0 flgOpt ! ;
: AGAIN   ( a -- )     xtBRAN  @ T, 1- Txt,     0 flgOpt ! ;
: UNTIL   ( a -- )     xtQBRAN @ T, 1- Txt,     0 flgOpt ! ;
: IF      ( -- A )     xtQBRAN @ T, HERE' 0 T,  0 flgOpt ! ;
: AHEAD   ( -- A )     xtBRAN  @ T, HERE' 0 T,  0 flgOpt ! ;
: THEN    ( A -- )     HERE' 1- SWAP Txt!       0 flgOpt ! ;
: ELSE    ( A1 -- A2 ) AHEAD  SWAP  THEN        0 flgOpt ! ;
: WHILE   ( a -- A a ) IF SWAP                  0 flgOpt ! ;
: REPEAT  ( A a -- )   AGAIN THEN               0 flgOpt ! ;


\ Токены для компиляции примитивов и определения СВУ
\ Взяты из таблиц, размещённых в eforth0.mkl
  0 prime N0				\   0 ||		1: начало
  1 prime STORE				\   1 !
  2 high ~DIG				\   2 #		СВУ
  3 high ~TICK				\   3 '		СВУ
  4 high ~PAREN		IMMEDIATE	\   4 (		СВУ
  5 prime STAR				\   5 *
  6 prime PLUS				\   6 +
  7 high ~COMMA				\   7 ,		СВУ
  8 prime SUBBB				\   8 -
  9 high ~DOT				\   9 .		СВУ
 10 prime SLASH				\  10 /
 11 prime ZERO				\  11 0
 12 high ~COLON				\  12 :		СВУ
 13 high ~SEMIS  IMMEDIATE COMPILE-ONLY	\  13 ;		СВУ
 14 prime LESS				\  14 <
 15 prime EQUAL				\  15 =
 16 high ~QUEST				\  16 ?		СВУ
 17 prime AT				\  17 @
 18 prime LBRAC		IMMEDIATE	\  18 [		СВУ
 19 prime BKSLA		IMMEDIATE	\  19 \
 20 prime RBRAC				\  20 ]		СВУ

 21 high ~EDIGS				\  21 #>	СВУ	2: начало
 22 high ~DIGS				\  22 #S	СВУ
 23 high ~STRQ		IMMEDIATE	\  23 $"	СВУ
 24 prime STASL				\  24 */
 25 prime PSTOR				\  25 +!
 26 high ~DOTQ		IMMEDIATE	\  26 ."	СВУ
 27 high ~DOTPR		IMMEDIATE	\  27 .(	СВУ
 28 high ~DOTR				\  28 .R	СВУ
 29 high ~DOTS				\  29 .s	СВУ
 30 prime ZSTORE			\  30 0!
 31 prime ZLESS				\  31 0<
 32 prime ONEP				\  32 1+
 33 prime ONEM				\  33 1-
 34 prime DSTOR				\  34 2!
 35 prime CELLS				\  35 2*
 36 prime CELLP				\  36 2+
 37 prime CELLM				\  37 2-
 38 prime CELLSL			\  38 2/
 39 prime DAT				\  39 2@
 40 high ~BDIGS				\  40 <#	СВУ
 41 prime TOR		COMPILE-ONLY	\  41 >R
 42 prime BLANK				\  42 BL
 43 prime CSTOR				\  43 C!
 44 high ~CCOMM				\  44 C,	СВУ
 45 prime CAT				\  45 C@
 46 prime CR				\  46 CR
 47 prime DPLUS				\  47 D+
 48 high ~IFF		IMMEDIATE	\  48 IF	СВУ
 49 prime MSTAR				\  49 M*
 50 prime ORR				\  50 OR
 51 prime RFROM				\  51 R>
 52 prime RAT				\  52 R@
 53 high ~UDOT				\  53 U.	СВУ
 54 prime ULESS				\  54 U<
 55 prime UAT				\  55 U@
 56 high ~BKSP				\  56 ^h	СВУ
 57 prime CP				\  57 cp

 58 prime STOIO				\  58 !io		3: начало
 59 high ~STRQP		COMPILE-ONLY	\  59 $"|	СВУ
 60 high ~STRCQ				\  60 $,"	СВУ
 61 high ~SNAME				\  61 $,n	СВУ
 62 high ~DOTQP		COMPILE-ONLY	\  62 ."|	СВУ
 63 high ~DOTID				\  63 .id	СВУ
 64 high ~DOTOK				\  64 .ok	СВУ
 65 prime INN				\  65 >in
 66 prime TIO				\  66 >io
 67 prime QRX				\  67 ?rx
 68 prime ABSS				\  68 ABS
 69 high ~AFT		IMMEDIATE	\  69 AFT	СВУ
 70 prime ANDD				\  70 AND
 71 prime BYE           		\  71 BYE
 72 prime DUPP				\  72 DUP
 73 high ~FORR		IMMEDIATE	\  73 FOR	СВУ
 74 prime HEX				\  74 HEX
 75 prime KEY				\  75 KEY
 76 prime MAX				\  76 MAX
 77 prime MIN				\  77 MIN
 78 prime MODD				\  78 MOD
 79 prime INVER				\  79 NOT
 80 prime PAD				\  80 PAD
 81 prime ROT				\  81 ROT
 82 high ~SEE				\  82 SEE	СВУ
 83 prime TIB				\  83 TIB
 84 high ~UDOTR				\  84 U.R	СВУ
 85 prime UMSTA				\  85 UM*
 86 prime UPLUS				\  86 UM+
 87 prime XORR				\  87 XOR
 88 high ~DOSTR		COMPILE-ONLY	\  88 do$	СВУ
 89 prime HLD				\  89 hld
 90 prime IOT				\  90 io>
 91 prime RPSTO		COMPILE-ONLY	\  91 rp!
 92 prime RZERO				\  92 rp0
 93 prime RPAT				\  93 rp@
 94 prime RXAT				\  94 rx@
 95 prime SPSTO				\  95 sp!
 96 prime SZERO				\  96 sp0
 97 prime SPAT				\  97 sp@
 98 high ~STRR				\  98 str	СВУ
 99 high ~TAP				\  99 tap	СВУ
100 prime TXSTO				\ 100 tx!
101 prime STOREA			\ 101 xt!
102 prime FROMXT			\ 102 xt>
103 prime AAT				\ 103 xt@

104 prime NTIB				\ 104 #tib		4: начало
105 prime TTIB				\ 105 'tib
106 prime SLMOD				\ 106 /MOD
107 prime DDUP				\ 107 2DUP
108 prime BBIT				\ 108 8bit
109 prime QDUP				\ 109 ?DUP
110 prime QKEY				\ 110 ?KEY
111 prime DROP				\ 111 DROP
112 high ~DUMP				\ 112 DUMP	СВУ
113 high ~ELSEE		IMMEDIATE	\ 113 ELSE	СВУ
114 prime EMIT				\ 114 EMIT
115 prime EXITT				\ 115 EXIT
116 prime FILL				\ 116 FILL
117 high ~FIND				\ 117 FIND	СВУ
118 prime HERE				\ 118 HERE
119 high ~HOLD				\ 119 HOLD	СВУ
120 high ~NEXT		IMMEDIATE	\ 120 NEXT	СВУ
121 prime OVER				\ 121 OVER
122 prime PICK				\ 122 PICK
123 high ~SIGN				\ 123 SIGN	СВУ
124 prime SWAP				\ 124 SWAP
125 high ~THENN		IMMEDIATE	\ 125 THEN	СВУ
126 prime TYPES				\ 126 TYPE
127 high ~WORDD				\ 127 WORD	СВУ
128 prime BASE				\ 128 base
129 high ~COLD				\ 129 cold	СВУ
130 prime CONT				\ 130 con>
131 high ~EVAL				\ 131 eval	СВУ
132 high ~FILE				\ 132 file	СВУ
133 high ~KTAP				\ 133 ktap	СВУ
134 prime LAST				\ 134 last
135 prime NGET				\ 135 nget
136 high ~NUFQ				\ 136 nuf?	СВУ
137 high ~QUIT				\ 137 quit	СВУ
138 prime WORK				\ 138 work

139 prime TBOOT				\ 139 'boot		5: начало
140 prime TEVAL				\ 140 'eval
141 prime SSMOD				\ 141 */MOD
142 prime DDROP				\ 142 2DROP
143 prime TCHAR				\ 143 >char
144 prime TNAME				\ 144 >name
145 high ~ABORT				\ 145 ABORT	СВУ
146 high ~AGAIN		IMMEDIATE	\ 146 AGAIN	СВУ
147 high ~ALLOT				\ 147 ALLOT	СВУ
148 high ~BEGIN		IMMEDIATE	\ 148 BEGIN	СВУ
149 prime CMOVEE			\ 149 CMOVE
150 prime COUNT				\ 150 COUNT
151 prime ERASE				\ 151 ERASE
152 high ~HELPR				\ 152 HELPR	СВУ
153 prime MSMOD				\ 153 M/MOD
154 high ~SPACE				\ 154 SPACE	СВУ
155 high ~UNTIL		IMMEDIATE	\ 155 UNTIL	СВУ
156 high ~WHILEE	IMMEDIATE	\ 156 WHILE	СВУ
157 high ~WORDS				\ 157 WORDS	СВУ
158 high ~AHEAD		IMMEDIATE	\ 158 ahead	СВУ
159 prime DEPTH				\ 159 depth
160 prime DIGIT				\ 160 digit
161 prime DOLIT		COMPILE-ONLY	\ 161 dolit
162 prime DONXT		COMPILE-ONLY	\ 162 donxt
163 prime HEADS				\ 163 head!
164 prime HEADA				\ 164 head@
165 prime NAMET				\ 165 name>
166 high ~NAMEQ				\ 166 name?	СВУ
167 high ~NTYPE				\ 167 ntype	СВУ
168 prime PACKS				\ 168 pack$
169 high ~PARSE				\ 169 parse	СВУ
170 high ~QUERY				\ 170 query	СВУ
171 high ~TLOAD				\ 171 tload	СВУ
172 high ~TOKEN				\ 172 token	СВУ
173 prime TYSTO				\ 173 type!

174 prime IFIND				\ 174 (find)		6: начало
175 high ~QSTAC				\ 175 ?stack	СВУ
176 high ~ABRTQ		IMMEDIATE	\ 176 ABORT"	СВУ
177 high ~CREAT				\ 177 CREATE	СВУ
178 prime NEGAT				\ 178 NEGATE
179 high ~REPEA		IMMEDIATE	\ 179 REPEAT	СВУ
180 high ~SPACS				\ 180 SPACES	СВУ
181 prime UMMOD				\ 181 UM/MOD
182 prime WITHI				\ 182 WITHIN
183 high ~ACCEP				\ 183 accept	СВУ
184 prime BRAN		COMPILE-ONLY	\ 184 branch
185 prime DIGTQ				\ 185 digit?
186 prime DOLITM	COMPILE-ONLY	\ 186 dolitm
187 prime NSPLIT			\ 187 nsplit
188 prime PRESE				\ 188 preset
189 prime RQUERY			\ 189 rquery

190 prime PARS				\ 190 (parse)		7: начало
191 prime QBRAN		COMPILE-ONLY	\ 191 ?branch
192 high ~UNIQU				\ 192 ?unique	СВУ
193 prime CMPR				\ 193 COMPARE
194 prime DECIM				\ 194 DECIMAL
195 prime DNEGA				\ 195 DNEGATE
196 prime EXECU				\ 196 EXECUTE
197 high ~ABORQ		COMPILE-ONLY	\ 197 abort"|	СВУ
198 high ~COMPI				\ 198 compile	СВУ
199 prime EXTRC				\ 199 extract
200 high ~LITER		IMMEDIATE	\ 200 literal	СВУ
201 high ~NUMBQ				\ 201 number?	СВУ
202 prime PUBLI				\ 202 publish

203 high ~SCOMP				\ 203 $compile	СВУ	8: начало
204 high ~ATEXE				\ 204 @EXECUTE	СВУ
205 prime COMPC				\ 205 COMPILE,
206 high ~CONST				\ 206 CONSTANT	СВУ
207 high ~VARIA				\ 207 VARIABLE	СВУ

208 prime IMMED				\ 208 IMMEDIATE		9: начало
209 high ~BCOMP		IMMEDIATE	\ 209 [compile]	СВУ

210 high ~INTER				\ 210 $interpret СВУ	10: начало


115 xtEXIT !    \ токен слова EXIT
184 xtBRAN  !	\ токен слова branch
191 xtQBRAN !	\ токен слова ?branch
211 nNames  !	\ полное число имён, включая нулевое имя

.( Tokens: ) nNames @ .         \ токены успешно определены


\ Определения высокого уровня
\ Обычные слова Форта. Векторное исполнение: @EXECUTE
\ E @EXECUTE   ( a -- )                 Исполнить вектор, хранящийся по адресу a.
~ATEXE :: ATEXE   UAT QDUP IF EXECU THEN EXITT ;;


\ Интерпретатор текста. Форматирование чисел: <# HOLD # #S SIGN #>
\ E <#         ( -- )                   Начать процесс числового вывода.
~BDIGS :: BDIGS   PAD HLD STORE EXITT ;; \ указать HLD на PAD, чтобы принять цифры числа

\ E HOLD       ( c -- )                 Вставить литеру в строку числового вывода.
~HOLD :: HOLD   HLD UAT ONEM	\ уменьшить HLD
                DUPP HLD STORE	\ цифры хранятся в обратном порядке
                CSTOR EXITT ;;	\ сохранить цифру c

\ E #          ( u -- u )               Извлечь одну цифру из u и добавить эту цифру в строку вывода.
~DIG :: DIG   BASE UAT EXTRC	\ извлечь младшую значащую цифру
              HOLD EXITT ;;	\ добавить её к строке числового вывода

\ E #S         ( u -- 0 )               Преобразовать u пока все цифры не будут добавлены в строку вывода.
~DIGS :: DIGS   BEGIN DIG DUPP	\ добавить одну цифру
                WHILE           \ если u не 0,
                REPEAT          \ повторять, пока u не сведётся к 0
                EXITT ;;

\ E SIGN       ( n -- )                 Добавить знак минуса к строке числового вывода.
~SIGN :: SIGN   ZLESS IF	\ если n<0,
                DOLIT 45 T, HOLD THEN	\ добавить знак минуса
                EXITT ;;

\ E #>         ( w -- b u )             Приготовить строку вывода к TYPE.
~EDIGS :: EDIGS   DROP HLD UAT	\ заменить w на HLD
                  PAD OVER SUBBB \ вернуть длину строки
                  EXITT ;;


\ Интерпретатор текста. Основной ввод-вывод: nuf? SPACE SPACES
\ Примитивы: TYPE CR
\ E nuf?       ( -- t )                 Вернуть false если нет ввода, иначе пауза и если CR вернуть true.
~NUFQ :: NUFQ   QKEY DUPP	\ получена клавиша?
   IF				\ нет, вернуть флаг false
     DDROP KEY			\ Да. Взять клавишу.
     DOLIT CRR T, EQUAL		\ Это CR? Вернуть флаг.
   THEN EXITT ;;

\ E SPACE      ( -- )                   Послать пустую литеру на устройство вывода.
~SPACE :: SPACE   BLANK EMIT EXITT ;;	\ послать пробел

\ E SPACES     ( +n -- )                Послать n пробелов на устройство вывода.
~SPACS :: SPACS   ZERO MAX TOR	\ избежать отрицательных чисел
                  AHEAD
b: CHAR1            SPACE	\ послать один пробел
                  THEN
                  DONXT CHAR1	\ цикл назад
                  EXITT ;;


\ Интерпретатор текста. Вывод чисел: str .R U.R U. .
\ Примитивы: HEX DECIMAL
\ E str        ( w -- b u )             Преобразовать целое со знаком в числовую строку.
~STRR :: STRR   DUPP TOR ABSS		\ сохранить w для SIGN и заменить w на абсолютную величину
                BDIGS DIGS RFROM	\ извлечь все цифры из абсолютной величины
                SIGN EDIGS EXITT ;;	\ добавить знак, вернуть буфер и длину

\ E .R         ( w +n -- )              Вывести целое в поле из n колонок, выровненное вправо.
~DOTR :: DOTR   TOR STRR RFROM		\ сохранить ширину колонки +n и преобразовать w
                OVER SUBBB SPACS	\ отобразить пробелы для выравнивания вправо
                TYPES EXITT ;;		\ вывести числовую строку

\ *** а если на входе отрицательное число? Сделать >U16
\ E U.R        ( w +n -- )              Вывести беззнаковое целое в n колонок, выровненное вправо.
~UDOTR :: UDOTR  TOR BDIGS DIGS EDIGS   \ преобразовать беззнаковое целое w
                 RFROM OVER SUBBB SPACS \ добавить пробелов для выравнивания вправо
                 TYPES EXITT ;;         \ вывести числовую строку

\ *** а если на входе отрицательное число? Сделать >U16
\ E U.         ( u -- )                 Вывести беззнаковое целое в свободном формате.
~UDOT :: UDOT   BDIGS DIGS EDIGS	\ преобразовать беззнаковое целое u
                SPACE TYPES EXITT ;;	\ добавить пробел и вывести числовую строку

\ E .          ( w -- )                 Отобразить целое в свободном формате, преварив пробелом.
\ На МК-161 даже недесятичные числа бывают со знаком.
~DOT :: DOT  ( w -- )   STRR SPACE TYPES EXITT ;; \ да, отобразить целое со знаком


\ Интерпретатор текста. Числовой ввод: number?
\ Примитив: digit?
\ E number?    ( a -- n T | a F )       Преобразовать числовую строку в целое. Положить флаг на вершину стека.
\ *** для скорости можно реализовать в коде
~NUMBQ :: NUMBQ
   BASE UAT TOR                 \ сохранить текущее основание из BASE
   ZERO OVER COUNT              \ a 0 a+1 n --, получить длину строки
   OVER CAT                     \ получить первую цифру
   DOLIT 36 T, EQUAL            \ это $ для шестнадцатеричной системы?
   IF HEX SWAP ONEP		\ да, использовать шестнадцатеричную систему и подстроить строку
      SWAP ONEM THEN		\ a 0 a+2 n-1 --
   OVER CAT			\ взять следующую цифру
   DOLIT 45 T, EQUAL TOR	\ Эта литера - ? Сохранить флаг
   SWAP RAT SUBBB SWAP		\ a 0 b' n' --, подстроить адрес b
   RAT PLUS QDUP		\ a 0 b" n" n" --, подстроить счётчик n"
   IF
     ONEM TOR			\ полноценный счётчик, преобразовать строку
b: NUMQ2
     DUPP TOR CAT		\ сохранить адрес b и взять следующую цифру
     BASE UAT DIGTQ		\ преобразовать её в соответствии с текущим основанием
     IF				\ если это полноценая цифра, 
       SWAP BASE UAT STAR	\ помножить её на основание
       PLUS RFROM ONEP          \ добавить к сумме. Увеличить адрес b
       DONXT NUMQ2		\ цикл назад к преобразованию следующей цифры
       RAT SWAP DROP		\ полностью преобразовать строку. Взять знак
       IF NEGAT THEN		\ поменять знак суммы если флаг - истина
       SWAP			\ sum a --
     ELSE
       RFROM RFROM DDROP	\ если встретилась не-цифра
       DDROP ZERO		\ a 0 --, преобразование не удалось
     THEN
     DUPP			\ sum a a --, если удача; иначе a 0 0 --
   THEN				\ NUMQ6:
   RFROM DDROP			\ сбросить мусор
   RFROM BASE STORE		\ восстановить BASE
   EXITT ;;


\ Интерпретатор текста. Строковые литералы: do$ $"| ."|
\ E do$      C ( -- a )                 Вернуть адрес скомпилированной строки.
\ *** можно попробовать реализовать в коде
~DOSTR :: DOSTR RFROM           \ 1-й адрес возврата должен быть сохранён
                RFROM ONEP DUPP \ 2-й адрес возврата указывает на строку со счётчиком a
                COUNT PLUS ONEM \ адрес следующего токена после строкового литерала
                TOR SWAP        \ заменить 2-й адрес возврата
                TOR EXITT ;;    \ восстановить сохранённый 1-й адрес возврата

\ E $"|      C ( -- a )                 Рутина времени исполнения, скомпилированная $". Вернуть адрес скомпилированной строки.
~STRQP :: STRQP DOSTR EXITT ;;  \ вызвать do$

\ E ."|      C ( -- )                   Рутина времени исполнения .". Вывести скомпилированную строку.
~DOTQP :: DOTQP DOSTR COUNT TYPES EXITT ;; \ отобразить следующую строку


\ Интерпретатор текста. Паржинг: PARSE
\ Примитив: (parse)
\ E parse      ( c -- b u )             Сканировать входной поток и вернуть строку со счётчиком, ограниченную c.
~PARSE :: PARSE
   TOR  TTIB UAT  INN UAT  PLUS	\ указатель на текущий входой буфер для начала паржинга
   NTIB UAT  INN UAT  SUBBB	\ длина оставшейся строки в TIB
   RFROM PARS			\ разобрать желанную строку
   INN PSTOR  EXITT ;;		\ передвинуть указатель на конец строки


\ Интерпретатор текста. Слова паржинга: .( ( WORD token
\ Примитив \
\ E .(       I ( -- )                   Вывести последующий текст вплоть до следующей ).
~DOTPR :: DOTPR   DOLIT 41 T, PARSE TYPES \ паржинг до ) и отобразить разобранную строку
   EXITT ;;  \ IMMEDIATE

\ E (        I ( -- )                   Игнорировать последующий текст, вплоть до следующей ). Комментарий.
~PAREN :: PAREN   DOLIT 41 T, PARSE DDROP \ паржинг до ) и убрать разобранную строку
   EXITT ;;  \ IMMEDIATE

\ E WORD       ( c -- a )               Разобрать слово из входного потока и скопировать его в словарь кода.
~WORDD :: WORDD   PARSE		\ паржинг до c
   HERE PACKS EXITT ;;		\ упаковать разобранную строку в буфер HERE

\ E token      ( -- a )                 Разобрать слово из входного потока и скопировать его в словарь имён.
~TOKEN :: TOKEN   BLANK WORDD	\ разобрать следующую строку, ограниченную пробелами
   EXITT ;;			\ упаковать разобранную строку в буфер HERE


\ Итерпретатор текста. Поиск в словаре: name?
\ Примитивы: name> COMPARE
\ E name?      ( a -- xt nfa | a F )    Поиск строки в словаре.
~NAMEQ :: NAMEQ   IFIND DUPP  IF DROP HEADA DROP THEN  EXITT ;;


\ Интерпретатор текста. Ввод текста: ^h tap ktap accept query
\ E ^h         ( bot eot cur -- bot eot cur ) Курсор назад, на одну литеру.
~BKSP :: BKSP
  TOR OVER RFROM		\ bot eot bot cur --
  SWAP OVER SUBBB		\ bot=cur?
  IF DOTQP MSG_BKSP ONEM        \ back space: ^h BL BL ^h ^h
  THEN EXITT ;;			\ bot=cur, не делать пробел назад

\ E tap        ( bot eot cur c -- bot eot cur ) Принять и отобразить нажатие клавиши, передвинуть курсор.
~TAP :: TAP   DUPP EMIT		\ дублировать литеру и отобразить её
              OVER CSTOR ONEP	\ сохранить c в cur, увеличить cur
              EXITT ;;

\ E ktap       ( bot eot cur c -- bot eot cur ) Обработать нажатие клавиши, CR или BS.
~KTAP :: KTAP
  DUPP DOLIT CRR T, SUBBB	\ клавиша это возврат?
  IF
    DOLIT BKSPP T, SUBBB	\ клавиша это пробел назад?
    IF BLANK TAP EXITT THEN	\ ничто из этого, заменить на пробел
    BKSP EXITT			\ KTAP1: обработать пробел назад
  THEN
  DROP SWAP DROP DUPP EXITT ;;	\ KTAP2: обработать возврат каретки

\ E accept     ( a u0 -- a u )          Принять литеры в буфер ввода. Вернуть их настоящее число.
~ACCEP :: ACCEP
  OVER PLUS OVER		\ a a+u0 a --
  BEGIN				\ ACCP1:
    DDUP SUBBB			\ a+u0 = текущий указатель?
  WHILE				\ да, выйти
    KEY DUPP			\ нет, взять следующую литеру
    BLANK BBIT UAT WITHI	\ допустимая литера?
    IF TAP			\ да, принять её в буфер ввода
    ELSE KTAP THEN		\ нет, ACCP2: обработать управляющую литеру
  REPEAT			\ ACCP3: цикл за следующуей литерой
  DROP OVER SUBBB EXITT ;;	\ ACCP4: сделано, вернуть настоящую длину строки

\ E query      ( -- )                   Принять входной поток в буфер ввода терминала.
~QUERY :: QUERY
  TIB DOLIT 80 T, ACCEP		\ принять до 80 литер в TIB
  NTIB STORE TTIB STORE		\ сохранить настоящую длину строки в #TIB, её адрес в 'TIB
  INN ZSTORE EXITT ;;		\ инициализировать >IN

\ Инструменты. Дамп стека: .s
\ E .s         ( ... -- ... )           Отобразить содержимое стека данных.
~DOTS :: DOTS
  TOR TOR TOR			\ сохранить три элемента на стеке
  DUPP DOT RFROM		\ вывести 4-й элемент
  DUPP DOT RFROM		\ вывести 3-й элемент
  DUPP DOT RFROM		\ вывести 2-й элемент
  DUPP DOT			\ вывести 1-й элемент
  DOTQP MSG_PROMPT		\ вывести разделитель " > "
  EXITT ;;


\ Интерпретатор текста. Обработка ошибок: ABORT abort"| ?stack
\ E ABORT      ( -- )                   Сбросить стек данных и прыгнуть на QUIT.
~ABORT :: ABORT PRESE DOTS      \ также вывести стек
                [F FW#1 0 T, ;; \ QUIT

\ E abort"|  C ( f -- )                 Код времени исполнения для ABORT". Уйти на ABORT с сообщением.
~ABORQ :: ABORQ IF			\ проверить флаг
                  DOSTR			\ получить адрес строки
b: ABOR1          SPACE COUNT TYPES	\ отобразить строку ошибки
                  DOLIT 63 T, EMIT CR ABORT \ идти после строки ошибки, 63 это код "?"
                THEN
                DOSTR DROP EXITT ;;	\ сбросить строку ошибки

\ E ?stack     ( -- )                   ABORT если стек данных исчерпан.
~QSTAC :: QSTAC DEPTH ZLESS	\ проверять только исчерпание
                ABORQ		\ abort если истина
                MSG_UNDERF	\ " underflow "
                EXITT ;;

\ Интерпретатор текста: $interpret .ok eval QUIT tload
\ Примитивы: [ preset

HERE' CONSTANT ^INTER		\ адрес $interpret для 'EVAL

\ E $interpret ( a -- )                 Интерпретировать слово. Если не получилось, попробовать преобразовать его в целое.
~INTER :: INTER
  IFIND				\ слово определено?
  IF				\ Нет. Преобразовать его в число
    HEADA DOLIT COMPO T, ANDD	\ проверка бита лексикона "только для компиляции"
    ABORQ			\ если оно только для компиляции, ABORT
    MSG_COMPO			\ " compile only"
    DROP EXECU EXITT		\ иначе, исполнить определённое слово
  THEN
  NUMBQ				\ преобразовать в число
  QBRAN ABOR1			\ не число, ABORT
  EXITT ;;

\ E .ok        ( -- )                   Отобразить стек данных, только когда интерпретируем.
~DOTOK :: DOTOK CR DOLIT ^INTER T, \ 'EVAL содержит $INTERPRET?
                TEVAL UAT EQUAL
                IF              \ нет, выйти
                  DOTS          \ да, вывести стек
                THEN EXITT ;;

\ E eval       ( -- )                   Интерпретировать входной поток.
~EVAL :: EVAL   BEGIN TOKEN DUPP CAT \ входной поток пуст?
                WHILE                \ да, выйти
                  TEVAL ATEXE QSTAC  \ нет, eval ввод, проверить стек
                REPEAT		\ цикл назад, за следующим словом
                DROP EXITT ;;	\ сделано

\ E quit       ( -- )                   Сбросить указатель стека и начать интерпретацию текста.
~QUIT :: QUIT   RZERO UAT RPSTO \ инициализировать указатель стека возвратов
                LBRAC           \ начать интерпретацию
                BEGIN QUERY     \ получить ввод
                  EVAL DOTOK    \ обработать ввод, вывести приглашение
                AGAIN ;;        \ продолжать до ошибки
FW#1 QUIT F]                    \ ABORT ссылался на QUIT, разрешить эту ссылку

\ R file       ( -- )                   Принять поток из универсального последовательного порта.
~FILE :: FILE
  RZERO UAT RPSTO		\ инициализировать указатель стека возвратов
  STOIO				\ инициализировать УПП
  CONT				\ терминал на консоль, для вывода диагностики ошибок
  LBRAC				\ начать интерпретацию
  BEGIN  RQUERY			\ получить ввод
         EVAL			\ обработать ввод, не выводить приглашение
         DOLIT 11 T, TXSTO	\ PACE означает успешную обработку принятой строки
  AGAIN ;;			\ продолжать до QUIT или ошибки


HERE' CONSTANT ^TLOAD		\ сохранить токен TLOAD для 'BOOT

\ R tload      ( -- )                   Интерпретировать область текста, как одну строку.
~TLOAD :: TLOAD
  DOLIT 5096 T, TTIB STORE	\ сделать область текста входным буфером
  DOLIT 3072 T, NTIB STORE	\ записать длину текста в #TIB
  INN ZSTORE			\ записать 0 в >IN
  LBRAC EVAL EXITT ;;		\ обработать текст


\ Компилятор двоеточием. Инструменты компилятора: ' ALLOT , C, compile [compile] literal $,"
\ Примитив COMPILE,
\ E '          ( -- xt )                Искать в словаре следующее слово из входного потока.
~TICK :: TICK   TOKEN NAMEQ     \ слово определено?
                QBRAN ABOR1
                EXITT ;;        \ да, адрес поля кода на стек

\ E ALLOT      ( n -- )                 Выделить n байт в словаре кода.
~ALLOT :: ALLOT CP PSTOR EXITT ;; \ подровнять указатель кода

\ E ,          ( w -- )                 Скомпилировать целое в словарь кода.
~COMMA :: COMMA
  HERE DUPP CELLP CP STORE      \ передвинуть CP
  STORE EXITT ;;                \ скомпилировать w в словарь

\ A C,         ( c -- )                 Скомпилировать байт в словарь кода.
~CCOMM :: CCOMM
  HERE DUPP ONEP CP STORE       \ передвинуть CP
  CSTOR EXITT ;;                \ скомпилировать c в словарь

\ E compile  C ( -- )                   Скомпилировать в словарь кода следующий адрес в списке двоеточия.
~COMPI :: COMPI
  RFROM DUPP ONEP UAT COMMA     \ скомпилировать адрес
  CELLP TOR EXITT ;;            \ подровнять адрес возврата

\ E [compile] I( -- ; <строка> )        Скомпилировать следующее слово IMMEDIATE в словарь кода.
~BCOMP :: BCOMP TICK COMPC EXITT ;; \ искать следующее слово и скомпилировать его xt
\ IMMEDIATE

\ E literal  I ( w -- )                 Скомпилировать вершину стека в словарь, как целый литерал.
~LITER :: LITER
  DUPP ZLESS  IF   COMPI DOLITM NEGAT \ если w<0, скомпилировать DOLITM и поменять знак w
              ELSE COMPI DOLIT THEN   \ если w>=0, скомпилировать DOLIT
  COMMA                         \ скомпилировать целый литерал
  EXITT ;;                      \ это целый литерал в слове двоеточия
\ IMMEDIATE

\ E $,"        ( -- )                   Скомпилировать строковый литерал вплоть до следующей ".
~STRCQ :: STRCQ
  DOLIT 34 T, PARSE HERE PACKS  \ скомпилировать строку в словарь кода
  COUNT PLUS                    \ подсчитать выровненный конец строки
  CP STORE EXITT ;;             \ подровнять указатель кода


\ Компилятор двоеточием. Стуктуры управления: FOR NEXT BEGIN UNTIL AGAIN IF ahead REPEAT THEN AFT ELSE WHILE
\ E FOR      I ( -- a )                 Начать структуру цикла FOR-NEXT в определении через двоеточия.
~FORR :: FORR   COMPI TOR       \ скомпилировать >R для начала цикла FOR-NEXT
                HERE EXITT ;;   \ оставить адрес a следующего токена
\ IMMEDIATE

\ E NEXT     I ( a -- )                 Завершить структуру цикла FOR-NEXT.
~NEXT :: NEXT  COMPI DONXT ONEM COMPC	\ скомпилировать адресный литерал DONXT с адресом a
               EXITT ;; \ IMMEDIATE

\ E BEGIN    I ( -- a )                 Начать структуру бесконечного или неопределённого цикла.
\ *** можем сделать синонимом к токену HERE, повлияет на кросс?
~BEGIN :: BEGIN   HERE EXITT ;;	\ оставить адрес следующего токена
\ IMMEDIATE

\ E UNTIL    I ( a -- )                 Завершить структуру неопределённого цикла BEGIN-UNTIL.
~UNTIL :: UNTIL   COMPI QBRAN ONEM COMPC \ скомпилировать адресный литерал ?BRANCH с адресом a
                  EXITT ;; \ IMMEDIATE

\ E AGAIN    I ( a -- )                 Завершить структуру бесконечного цикла BEGIN-AGAIN.
~AGAIN :: AGAIN   COMPI BRAN ONEM COMPC	\ скомпилировать адресный литерал BRANCH с адресом a
                  EXITT ;; \ IMMEDIATE

\ E IF       I ( -- A )                 Начать структуру условного ветвления.
~IFF :: IFF       COMPI QBRAN HERE	\ скомпилировать адресный литерал ?BRANCH, оставить адрес A
                  ZERO COMMA		\ инициализировать поле адреса нулём
                  EXITT ;; \ IMMEDIATE

\ E ahead    I ( -- A )                 Скомпилировать инструкцию ветвления вперёд.
~AHEAD :: AHEAD   COMPI BRAN HERE	\ скомпилировать адресный литерал BRANCH, оставить адрес A
                  ZERO COMMA		\ инициализировать поле адреса нулём
                  EXITT ;; \ IMMEDIATE

\ E THEN     I ( A -- )                 Завершить структуру условного ветвления.
~THENN :: THENN   HERE ONEM SWAP STOREA	\ разрешить адрес в A адресом текущего токена
                  EXITT ;; \ IMMEDIATE

\ E REPEAT   I ( A a -- )               Завершить неопределённый цикл BEGIN-WHILE-REPEAT.
~REPEA :: REPEA   AGAIN		\ скомпилировать адресный литерал BRANCH с адресом a
                  THENN		\ разрешить адрес в A адресом текущего токена
                  EXITT ;; \ IMMEDIATE

\ E AFT      I ( a -- a A )             Прыгнуть к THEN в цикле FOR-AFT-THEN-NEXT на первом проходе.
~AFT :: AFT   DROP AHEAD		\ скомпилировать адресный литерал BRANCH и оставить A
              BEGIN SWAP EXITT ;;	\ поменять a с адресом текущего токена
\ IMMEDIATE

\ E ELSE     I ( A -- A )               Начать ложное условие в структуре IF-ELSE-THEN.
~ELSEE :: ELSEE   AHEAD SWAP		\ скомпилировать адресный литерал BRANCH. Разрешить адрес в A
                  THENN EXITT ;;	\ адресом текущего токена. Заменить A адресом литерала
\ IMMEDIATE

\ E WHILE    I ( a -- A a )             Условное ветвление из цикла BEGIN-WHILE-REPEAT.
~WHILEE :: WHILEE   IFF SWAP EXITT ;;	\ скомпилировать адресный литерал BRANCH. Оставить адрес A литерала
\ IMMEDIATE


\ Компилятор двоеточием. Строковые слова: ABORT" $" ."
\ E ABORT"   I ( -- ; <строка> )        Начать структуру условного ветвления.
~ABRTQ :: ABRTQ  COMPI ABORQ STRCQ	\ скомпилировать строковый литерал abort"| с последующей строкой
                 EXITT ;; \ IMMEDIATE

\ E $"       I ( -- ; <строка> )        Скомпилировать вшитый строковый литерал.
~STRQ :: STRQ   COMPI STRQP STRCQ	\ скомпилировать строковый литерал $"| с последующей строкой
                EXITT ;; \ IMMEDIATE

\ E ."       I ( -- ; <строка> )        Скомпилировать вшитый строковый литерал для вывода во время исполнения.
~DOTQ :: DOTQ   COMPI DOTQP STRCQ	\ скомпилировать строковый литерал ."| с последующей строкой
                EXITT ;; \ IMMEDIATE


\ Компилятор двоеточием. Поле имени: ?unique $,n
\ E ?unique    ( a -- a )               Отобразить предупреждение, если слово уже существует.
~UNIQU :: UNIQU   DUPP IFIND		\ слово определено?
                  IF			\ переопределения разрешены,
                    DOTQP 		\ но пользователь должен быть предупреждён
                    MSG_REDEF		\ " reDef "
                    OVER COUNT TYPES	\ просто на случай, если это непреднамеренно
                  ELSE DROP ZERO	\ если слова с таким именем не было, сохранить 0
                  THEN			\ UNIQ1:
                  LAST STORE EXITT ;;	\ LAST := номер найденного регистра заголовка

\ E $,n        ( nfa -- )               Построить новый заголовок для словаря, используя строку в nfa.
~SNAME :: SNAME
  DUPP CAT			\ пустой ввод?
  IF
    UNIQU			\ повторное определение?
    LAST UAT QDUP		\ старый заголовок был найден?
    IF   HEADA			\ да, был найден: ( xt nfa lex )
         ROT DDROP HERE SWAP	\ ( cp, старое NFA )
         WORK HEADS		\ построить новый заголовок в WORK
         DROP EXITT		\ убрать новое nfa из стека
    ELSE DUPP COUNT PLUS	\ пропустить поле имени
         DUPP CP STORE		\ теперь CP указывает на поле кода
         SWAP WORK HEADS EXITT	\ построить заголовок с новым полем имени в WORK
    THEN
  THEN				\ PNAM1:
  STRQP MSG_NAME		\ предупреждение " name"
  BRAN ABOR1 ;;			\ пустой ввод

\ Компилятор двоеточием. Компилятор Форта: $compile ; :
\ Примитивы: ] publish IMMEDIATE

HERE' CONSTANT ^SCOMP		\ Адрес $compile для 'EVAL

\ E $compile   ( a -- )                 Скомпилировать следующее слово в словарь кода, как токен или литерал.
~SCOMP :: SCOMP
  IFIND				\ слово определено?
  IF
    HEADA DOLIT IMEDD T, ANDD	\ IMMEDIATE ?
    IF DROP EXECU EXITT THEN	\ это IMMEDIATE, исполнить
    DROP COMPC EXITT		\ это не IMMEDIATE, скомпилировать
  THEN
  NUMBQ QBRAN ABOR1		\ попробовать преобразовать в число
  LITER EXITT ;;		\ скомпилировать число, как целый литерал

\ E ;       IC ( -- )                   Завершить определение двоеточием.
~SEMIS :: SEMIS  COMPI EXITT    \ скомпилировать EXIT
  LBRAC PUBLI EXITT ;;          \ вернуться в режим интерпретации
\ IMMEDIATE COMPILE-ONLY

\ E :          ( -- ; <строка> )        Начать новое определение двоеточием, используя следующее слово, как его имя.
~COLON :: COLON  TOKEN SNAME    \ получить новую строку и построить новое поле имени
  DOLIT doLst' @ T, CCOMM       \ скомпилировать DOLST в поле кода
  RBRAC EXITT ;;                \ переключиться в режим компиляции


\ Компилятор двоеточием. Определяющие слова: CREATE CONSTANT VARIABLE
\ E CREATE     ( -- ; <строка> )        Скомпилировать новый вход массива без выделения пространства кода.
~CREAT :: CREAT  TOKEN SNAME PUBLI \ построить новый заголовок
  DOLIT doVar' @ T, CCOMM EXITT ;; \ скомпилировать DOVAR в поле кода

\ E CONSTANT   ( n -- ; <строка> )      Скомпилировать новую константу.
~CONST :: CONST  TOKEN SNAME PUBLI  \ построить новый заголовок
  DUPP ZLESS IF  NEGAT              \ если n<0, изменить знак и
                 DOLIT doConM' @ T, \ скомпилировать DOCONM в поле кода (отрицательная константа)
             ELSE DOLIT doCon' @ T, \ скомпилировать DOCON в поле кода (положительная константа)
             THEN
  CCOMM COMMA EXITT ;;		\ скомпилировать n в поле параметров

\ E VARIABLE   ( -- ; <строка> )        Скомпилировать новую переменную, инициализировав нулём.
~VARIA :: VARIA  CREAT		\ скомпилировать новый заголовок и поле кода DOVAR
  ZERO COMMA EXITT ;;		\ инициализировать поле параметров нулём


\ -----	Последующие слова не используются eForth и могут быть скомпилированы позже, уже на борту МК-161
.( Core: ) HERE' . CR           \ Вывести размер ядра

\ Поиск в словаре из ANS: FIND
\ A FIND       ( a -- a F | xt 1 | xt -1) Поиск строки в словаре. Вернуть 1, если IMMEDIATE.
~FIND :: FIND  IFIND DUPP
  IF DROP HEADA DOLIT IMEDD T, ANDD	\ если найден, проверить IMMEDIATE
     IF DROP DOLIT 1 T, EXITT THEN	\ если IMMEDIATE, вернуть ( xt 1 )
     DROP DOLITM 1 T,			\ иначе вернуть ( xt -1 )
  THEN
  EXITT ;;

\ Вывод чисел: ?
\ E ?          ( a -- )                 Отобразить содержимое ячейки памяти.
~QUEST :: QUEST   AT DOT EXITT ;;


\ Инструменты: DUMP .id WORDS SEE
\ Примитивы: >name
\ E DUMP       ( b -- )                 Вывести 20 байт из b, форматированным способом.
~DUMP :: DUMP
  DUPP DOLIT 1 T, DOLIT 1000 T, WITHI \ вывести область десятичных регистров?
  IF DOLIT 4 T, TOR		\ вывести 5 регистров
b: DUMPD
       CR DUPP DOLIT 3 T, UDOTR	\ вывести номер десятичного регистра
       SPACE COUNT DOT		\ вывести пробел и содержимое регистра
     DONXT DUMPD  DROP EXITT
  THEN
  BASE UAT DOLIT 16 T, EQUAL	\ шестнадцатеричный дамп?
  IF DOLIT 3 T, TOR		\ четыре форматированные строки
b: DUMPH
       DUPP CR
       DOLIT 1 T, TOR		\ повторить два раза
b: DUMPH1
         COUNT DOLIT 2 T, UDOTR	\ вывести 2 байта
         COUNT DOLIT 3 T, UDOTR
         DOLIT TABB T, EMIT	\ табуляция
       DONXT DUMPH1		\ итого выведено 4 байта
       COUNT DOLIT 2 T, UDOTR	\ вывести 5-й байт
       SPACE SPACE		\ два пробела
       SWAP DOLIT 5 T, TYPES	\ вывести 5 литер
     DONXT DUMPH  DROP EXITT
  THEN				\ обычный (десятичный) дамп:
  DOLIT 1 T, TOR		\ в десятичном дампе 2 строки
b: DUMP1
     DUPP CR DUPP DOLIT 1 T, DOTR SPACE \ вывести адрес и пробел
     DOLIT 9 T, TOR		\ вывести 10 байт данных
b: DUMP2
       COUNT DOT
       DONXT DUMP2
     SPACE SWAP	DOLIT 10 T, TYPES \ добавить пробел, вывести 10 байт текста
  DONXT DUMP1  DROP EXITT ;;

\ E .id        ( nfa -- )               Отобразить имя по адресу поля имени.
~DOTID :: DOTID
  QDUP IF			\ если ноль, нет имени
         COUNT TYPES EXITT	\ отобразить строку имени
       THEN
  DOTQP MSG_NONAME		\ нет имени " {noName}"
  EXITT ;;

\ E WORDS      ( -- )                   Отобразить имена в контекстном словаре.
~WORDS :: WORDS
  DOLIT 300 T,			\ Часть 1. Начнать с регистра 300+1
  DOLIT 31 T, TOR		\ повторить 32 раза, для регистров 301..332
b: WORS1
    ONEP DUPP HEADA		\ ( reg max r0 num )
    DUPP IF			\ если слова в списке есть
      DDUP ONEM TOR		\ ( reg max r0 num r0 ), повторить num раз
b: WORS2
        DUPP HEADA DROP SPACE DOTID DROP \ отобразить имя
        NUFQ			\ управление пользователя
        IF RFROM RFROM DDROP	\ выход из циклов
           DDROP DDROP DROP EXITT \ очистка стека, остановка пользователем
        THEN
        ONEP			\ ( reg max r0 num r0+1 )
      DONXT WORS2  DROP		\ перейти к следующему регистру в списке
    THEN  DROP DDROP		\ ( reg ), список обработан
  DONXT WORS1  DROP		\ перейти к следующему списку слов
  DOLIT 44 T,			\ Часть 2. Начнём с регистра 44+1
  DOLIT nNames @ 2- T, TOR	\ отобразить все встроенные имена, кроме 0
b: WORS3
    ONEP DUPP HEADA DROP SPACE DOTID DROP \ отобразить имя
    NUFQ			\ управление пользователя
    IF RFROM DDROP EXITT THEN	\ остановка пользователем
  DONXT WORS3			\ повторить следующее слово
  DROP EXITT ;;

\ E SEE        ( -- ; <строка> )        Простой декомпилятор. Обновлён для байтовых машин, 08мар98cht
~SEE :: SEE  TICK CR		\ начальный адрес
b: SEE1
    ONEP DUPP AAT DUPP		\ он содержит ноль?
    IF TNAME THEN		\ это имя?
    QDUP			\ адрес имени или ноль
    IF SPACE DOTID		\ отобразить имя
       ONEP			\ следующий токен
    ELSE DUPP CAT UDOT		\ отобразить число
    THEN
    NUFQ			\ управление пользователя
  QBRAN SEE1			\ декомпилировать следующий токен
  DROP EXITT ;;


\ Справка на борту: ntype HELPR
\ Примитивы: nsplit nget
\ R ntype      ( a l -- )               Вывести строку длиной l, начиная с записи блокнота a.
~NTYPE :: NTYPE
  HERE NGET TYPES EXITT ;;      \ считать строку из блокнота в HERE, отобразить её.

\ R HELPR      ( n -- )                 Вывести справку по регистру n.
~HELPR :: HELPR
  CR DOLIT 82 T, EMIT		\ вывести 'R' на новой строчке
  DUPP STRR TYPES		\ вывести номер регистра
  DOLIT 58 T, EMIT SPACE	\ вывести ": "
  DOLIT 9000 T, SUBBB NSPLIT	\ считать адрес справки по регистру
  QDUP
  IF DOLIT 1000 T, PLUS  NSPLIT \ считать адрес справки по разделу
     DROP NTYPE CR		\ вывести справку по разделу
  THEN
  NTYPE EXITT ;;		\ вывести справку по регистру


\ Аппаратный сброс: cold
\ E cold       ( -- )                   Высокоуровневая последовательность холодного старта.
~COLD :: COLD  CONT                \ терминал на консоль
               DECIM               \ установить основание, на МК-161 десятичная адресация
               DOTQP MSG_SIGNON CR \ входное приветствие "161eForth v#.#"
               TBOOT ATEXE         \ интерпретировать текст
               ABORT ;;            \ начать интерпретацию


{ \ Начальные значения двоичных регистров, в файл MKB
  \ - системные переменные терминалов МК-161
   1         flgKbdCaps TC!     \ по умолчанию буквы заглавные

  \ - системные переменные eForth
HERE'        _CP        T!      \ отсюда начинается словарь пользователя
^TLOAD       _TBOOT     T!      \ загрузка из области текста

  \ Начальные значения десятичных регистров, в файл MKD:
INEXT' @      9         TR!     \ R9 указывает на адресный интерпретатор INEXT
 256         14         TR!     \ RE = 256
&KPNum @     ptrKbdInt  TR!     \ состояние клавиатуры: режим Num

  \ - точки входа в код высокого уровня
^INTER >3+3  rINTER     TR!     \ вектор $INTERPRET для [
^SCOMP >3+3  rCOMP      TR!     \ вектор $COMPILE для ]
~COLD  44 +  rCOLD      TR!     \ адрес заголовка COLD для автостарта шитого кода в MAIN
}

process.mkp                     \ построить таблицу заголовков и собрать ядро
.( HERE'=) HERE' .              \ краткая информация о собранном образе
.( wc=) wc . .( mc=) mc .       \ количество СВУ и примитивов
save.mkb  save.mkd  save.mkp    \ сохранить eForth
.( Saved!) CR                   \ вывести сообщение об успехе

ONLY FORTH BYE                  \ конец
\EOF
