
/* mkmkn.c v0.6.1b (c) Ilya V. Vasilyev, Moscow, 2019 */
/* 161eForth tool to create eforth.mkn                */
/* Distributed under GNU GPL v3 or later              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define REGIONS (30)
#define MAXMKN (524292)

FILE *fIn, *fOut;
static unsigned char line[512];             /* input line       */
static unsigned char text[65536];           /* registers buffer */
static unsigned char regions[REGIONS*1024]; /* regions buffer   */

static unsigned char mknImage[MAXMKN];      /* eforth.mkn image */

int tregs[1000], lregs[1000], rregs[1000], ptext;    /* describes R9000..R9999 */
int fregions[REGIONS], lregions[REGIONS], pregions;  /* register regions       */

int iRegion;                                /* active region number */


void Init(void)
{ int i;
  for (i=0; i<1000; i++)                    /* initialize register descriptions */
  { tregs[i]=0;
    lregs[i]=0;
    rregs[i]=0;
  }
  for (i=0; i<REGIONS; i++)
  { fregions[i]=0;                          /* intialize regions */
    lregions[i]=0;
  }
  iRegion = 0;                              /* start from region 1 */
  pregions = 0;                             /* start from regions[0] */
  ptext = 0;                                /* start from text[0] */
  for (i=0; i<MAXMKN; i++)                  /* zero the notebook image */
    mknImage[i]=0;
}


void Pack2Ch(unsigned char *s, int i1, int i2) /* pack 2 characters into 3 bytes */
{ s[0] = (i1/100)*16 + (i1/10)%10;
  s[1] = (i1%10)*16 + (i2/100);
  s[2] = ((i2/10)%10)*16 + (i2%10);
}


void PackDec(unsigned char *src, unsigned char *dest) /* pack 4 characters into decimal */
{ dest[0]=0;  /* positive */
  dest[7]=11; /* exponent */
  Pack2Ch(dest+1, src[0], src[1]);
  Pack2Ch(dest+4, src[2], src[3]);
}


void Normalize(unsigned char *s) /* normalize a decimal */
{ if ((s[1]!=0) || (s[2]!=0) || (s[3]!=0) ||
      (s[4]!=0) || (s[5]!=0) || (s[6]!=0)) /* if not 0 */
    while (s[1]<16)
    { s[1]=(((s[1]&15)<<4) | (s[2]>>4));
      s[2]=(((s[2]&15)<<4) | (s[3]>>4));
      s[3]=(((s[3]&15)<<4) | (s[4]>>4));
      s[4]=(((s[4]&15)<<4) | (s[5]>>4));
      s[5]=(((s[5]&15)<<4) | (s[6]>>4));
      s[6]= ((s[6]&15)<<4);
      (s[7])--;
    }
  else s[7]=0; /* if s==0 then exponent is also 0 */
}


int PackStr(unsigned char *src, unsigned char **d)
/* pack string to decimals */
{ static unsigned char str[4];
  int n=0, i=0;
  str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
  while(*src)
  { if ((*src!=13) && (*src!=10))
    { str[i++] = *src;  n++;
      if (i==4)
      { PackDec(str,*d);  Normalize(*d);  *d += 8;
        str[0]=0;  str[1]=0;  str[2]=0;  str[3]=0;
        i = 0;
      }
    }
    src++;
  }
  if (i)
  { PackDec(str,*d);  Normalize(*d);  *d += 8;
  }
  return n; /* length of packed string */
}


int IsBlank(unsigned char *s)
{ int i=0;
  unsigned char c=s[0];
  while (c)
  { if ((c!=32) && (c!=13) && (c!=10) && (c!=9)) return 0; /* string s not blank */
    c=s[i++];
  }
  return 1;                                /* string s is blank */
}


void StoreRegion(unsigned char *s)          /* save new region name */
{ unsigned char *p = regions + pregions;
  fregions[++iRegion] = pregions;
  lregions[iRegion] = PackStr(s,&p);
  pregions = p - regions;
}


void StoreRegister(unsigned char *s)        /* save register definition */
{ if (s[0]!='R') fprintf(stderr, "Wrong register: %s|", line);
  else
  { int n = (((s[1]-'0')*10+s[2]-'0')*10+s[3]-'0')*10+s[4]-'0'; /* register number */
    unsigned char *p = text + ptext;
    rregs[n-9000] = iRegion;
    tregs[n-9000] = ptext;
    lregs[n-9000] = PackStr(s+7,&p);
    ptext = p - text;
  }
}


void MkReg(int reg, int nn, int aaaaa, int lllll)
/* Rn := NNAAAAALLLLL */
{ unsigned char *p = mknImage+4+reg*8;
  p[0] = 0;  /* positive */
  p[7] = 11; /* default exponent */
  p[1] = ((nn/10)<<4) + (nn%10);
  p[2] = ((aaaaa/10000)<<4) + ((aaaaa/1000)%10);
  p[3] = (((aaaaa/100)%10)<<4) + ((aaaaa/10)%10);
  p[4] = ((aaaaa%10)<<4) + (lllll/10000);
  p[5] = (((lllll/1000)%10)<<4) + ((lllll/100)%10);
  p[6] = (((lllll/10)%10)<<4) + (lllll%10);
  Normalize(p); /* eliminate leading zeros */
}


void MkMKN(void)
/* pack parsed help text into mkn file */
{ int i;
  int size = 251 + (iRegion + (pregions+ptext)/8)/4;
  mknImage[0] = 0;                         /* starting from group 0 */
  mknImage[1] = 0;
  mknImage[2] = (size&255);                /* last group */
  mknImage[3] = (size>>8);

  for (i=0; i<1000; i++)                   /* make register descriptors */
    if (lregs[i]) /* if there is help line for that register */
      MkReg(i, rregs[i], 1001+iRegion+(pregions+tregs[i])/8, lregs[i]);

  for (i=1; i<=iRegion; i++)               /* make region descriptors */
    MkReg(1000+i, 0, 1001+iRegion+fregions[i]/8, lregions[i]);

  fwrite(mknImage, 1, 4, fOut);              /* write header */
  fwrite(mknImage+4, 8, 1001+iRegion, fOut); /* write descriptors */
  fwrite(regions, 8, pregions/8, fOut);      /* write region definitions */
  fwrite(text, 8, ptext/8, fOut);            /* write register definitions */

  size = (1 + iRegion + (pregions + ptext)/8)%4;
  if (size)
    fwrite(mknImage+4, 8, 4-size, fOut);     /* align last group */
}


int main(int arc, char *argv[])
{ fIn = fopen(argv[1],"rb");
  fprintf(stderr, "mkmkn v0.6.1b (c) Ilya V. Vasilyev\n"
                  "Distributed under GNU GPL v3 or higher\n");
  Init();
  if (fIn!=NULL)
  { while (fgets((char *)line, 512, fIn)!=NULL) /* parse help file */
      if (IsBlank(line)==0)
      { StoreRegion(line);                /* first non blank line is the region name */
        while (fgets((char *)line, 512, fIn)!=NULL)
        { if (IsBlank(line)) break;       /* first blank line ends the region block */
          StoreRegister(line);
        }
      }
    fclose(fIn);
    fprintf(stderr, "text = %d;  regions = %d\n", ptext/8, pregions/8);
    fOut = fopen(argv[2], "wb");
    if (fOut!=NULL)
    { MkMKN();
      fclose(fOut);
      return 1;
    } else fprintf(stderr, "Can't create file: %s\n", argv[2]);
  } else fprintf(stderr, "File not found: %s\n", argv[1]);
  fprintf(stderr, "Usage: mkmkn infile.txt outfile.mkn\n");
  return 0;
}
