
/* typemkn.c v0.6.1b (c) Ilya V. Vasilyev, Moscow, 2019 */
/* 161eForth tool to browse eforth.mkn                  */
/* Distributed under GNU GPL v3 or later                */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXMKN (524292)

FILE *fIn;
static unsigned char mknImage[MAXMKN];      /* eforth.mkn image */


void UnPack2Ch(unsigned char *s, unsigned char *c1, unsigned char *c2)
/* unpack 2 character from 3 bytes */
{ *c1 = (s[0]/16)*100 + (s[0]%16)*10 + (s[1]/16);
  *c2 = (s[1]%16)*100 + (s[2]/16)*10 + (s[2]%16);
}


void UnPackD(unsigned char *from, unsigned char *to)
/* unpack 4 characters from decimal */
{ UnPack2Ch(from+1, to, to+1);
  UnPack2Ch(from+4, to+2, to+3);
}


void UnNorm(unsigned char *s) /* force EE 11 */
{ if (s[7]!=0)
    while (s[7]<11)
    { s[6] = ( ((s[5]&15)<<4) | (s[6]>>4));
      s[5] = ( ((s[4]&15)<<4) | (s[5]>>4));
      s[4] = ( ((s[3]&15)<<4) | (s[4]>>4));
      s[3] = ( ((s[2]&15)<<4) | (s[3]>>4));
      s[2] = ( ((s[1]&15)<<4) | (s[2]>>4));
      s[1] = (s[1]>>4);
      (s[7])++;
    }
}


void NSplit(int n, int *prgn, int *padr, int *plen)
{ unsigned char *p = mknImage+4+n*8;
  UnNorm(p);
  *prgn = (p[1]>>4)*10 + (p[1]&15);
  *padr = (p[2]>>4)*10000 + (p[2]&15)*1000 + (p[3]>>4)*100 + (p[3]&15)*10 + (p[4]>>4);
  *plen = (p[4]&15)*10000 + (p[5]>>4)*1000 + (p[5]&15)*100 + (p[6]>>4)*10 + (p[6]&15);
}


void NTypeLn(unsigned char *s, int l)       /* print l characters from s, new line */
{ unsigned char str[4];
  int i;
  UnNorm(s);  UnPackD(s,str);  s+=8;  i=0;
  while (l)
  { putchar(str[i++]);
    if (i==4)
    { UnNorm(s);  UnPackD(s,str);  s+=8;  i=0;
    }
    l--;
  }
  puts("");
}


void TypeF(void)
{ int i=0;
  for(i=0; i<1000; i++)
  { static int region, nn, adr, len, adr2, len2;
    NSplit(i, &region, &adr, &len);
    if (region)
    { fprintf(stdout, "R%d: ", 9000+i);
      NSplit(1000+region, &nn, &adr2, &len2);
      NTypeLn(mknImage+4+adr2*8, len2);
      NTypeLn(mknImage+4+adr*8, len);
    }
  }
}


void ReadMKN(void)
{ int i;
  int size;
  fread(mknImage, 1, 4, fIn);              /* read header */
  if ((mknImage[0]!=0) || (mknImage[1]!=0))
  { fprintf(stderr, "mkn: first group is not zero\n");
    return;
  }
  size = (mknImage[3]<<8) + mknImage[2];    /* last group */
  fprintf(stderr, "Reading groups 0..%d\n", size);
  fread(mknImage+4, 8*4, size+1, fIn);     /* read all groups */
}


int main(int arc, char *argv[])
{ fIn = fopen(argv[1],"rb");
  fprintf(stderr, "typemkn v0.6.1b (c) Ilya V. Vasilyev\n"
                  "Distributed under GNU GPL v3 or higher\n");
  if (fIn!=NULL)
  { ReadMKN();
    fclose(fIn);
    TypeF();
  } else
  { fprintf(stderr, "File not found: %s\n", argv[1]);
    fprintf(stderr, "Usage: typemkn infile.mkn\n");
    return 0;
  }
  return 1;
}
